#pragma once

#include <Base_Provider.h>
#include <Base_Configurator.h>
#include <unordered_map>
#include <librdkafka/rdkafkacpp.h>
#include <atomic>
#include <thread>
#include <Encoder.hh>
#include <mutex>

namespace inaf::oasbo::Providers {

/**
 * @class KafkaAvroProvider
 * @brief The KafkaAvroProvider class is a derived class of BaseProvider and provides functionality for sending
 *        BasePackets to a Kafka topic in Avro format.
 * Check  Base_Provider for other information.
 */

class KafkaAvroProvider: public BaseProvider {
protected:

	//Kafka
	RdKafka::Producer *producer = nullptr; /**< Pointer to the Kafka producer object. */
	std::thread pollThread; /**< Thread for polling Kafka events. */
	bool pollThreadFlag = false; /**< Flag to indicate if the poll thread is running. */
	bool _open = false; /**< internal variable to check the state of the provider */
	class KafkaDeliveryReportCb; /**< Forward declaration of the KafkaDeliveryReportCb class. */
	KafkaDeliveryReportCb *dr_cb = nullptr; /**< Pointer to the Kafka delivery report callback object. */
	class KafkaConnectionCallback; /**< Forward declaration of the KafkaConnectionCallback class. */
	KafkaConnectionCallback *connection_cb = nullptr; /**< Pointer to the Kafka connection callback object. */

	// AVRO
	avro::EncoderPtr avroencoder; /**< Pointer to the AVRO encoder object. */

	KafkaAvroProvider(); /**< Default constructor. */
	KafkaAvroProvider(std::string ip, int port, std::string topic); /**< Constructor with parameters. */
	void _close(); /**< Performs the "privates" closing operations. */
	void flush(); /**< Performs the flush of queued packets */
	/**
	 * @brief Function for the polling thread.
	 */
	void pollingThreadFunction();

	/**
	 * @brief Function for encoding a BasePacket to AVRO format.
	 * @param packet The BasePacket to encode.
	 */
	void encodeToAvro(Packets::BasePacket&);

public:
	std::string brokerIp; /**< The IP address of the Kafka broker. */
	int brokerPort; /**< The port number of the Kafka broker. */

	void setDest(std::string dest) override;

	std::string getDest() override;

	int write(Packets::BasePacket&) override;

	int write(Packets::BasePacket&, std::string dest) override;

	int close() override;

	int open() override;

	bool isOpen() override;

	~KafkaAvroProvider(); /**< Destructor. */

	friend class KafkaAvroProviderBuilder;

};

/**
 * @class KafkaAvroProviderBuilder
 * @brief The KafkaAvroProviderBuilder class is used to build KafkaAvroProvider objects with configurable parameters.
 */
class KafkaAvroProviderBuilder {
protected:
	KafkaAvroProvider *provider;

public:
	std::string config_target { "kafkaavroprovider" };
	std::string ip_key { "ip" };
	std::string port_key { "port" };
	std::string topic_key { "topic" };

	KafkaAvroProviderBuilder();
	KafkaAvroProviderBuilder(std::string ip, int port, std::string topic);
	~KafkaAvroProviderBuilder();

	/**
	 * @brief Resets the builder to its initial state.
	 */
	void reset();

	/**
	 * @brief Configures the builder using a BaseConfigurator object.
	 * @param conf The BaseConfigurator object to configure from.
	 * @return A pointer to the KafkaAvroProviderBuilder.
	 */
	KafkaAvroProviderBuilder* configFrom(Configurators::BaseConfigurator &conf);

	/**
	 * @brief Sets the IP address for the KafkaAvroProvider.
	 * @param ip The IP address to set.
	 * @return A pointer to the KafkaAvroProviderBuilder.
	 */
	KafkaAvroProviderBuilder* setIp(std::string ip);

	/**
	 * @brief Sets the port number for the KafkaAvroProvider.
	 * @param port The port number to set.
	 * @return A pointer to the KafkaAvroProviderBuilder.
	 */
	KafkaAvroProviderBuilder* setPort(int port);

	/**
	 * @brief Sets the topic for the KafkaAvroProvider.
	 * @param topic The topic to set.
	 * @return A pointer to the KafkaAvroProviderBuilder.
	 */
	KafkaAvroProviderBuilder* setTopic(std::string topic);

	/**
	 * @brief Gets the built KafkaAvroProvider object.
	 * @return A pointer to the built KafkaAvroProvider.
	 */
	KafkaAvroProvider* getProvider();
};

} // namespace inaf::oasbo::Providers
