#include <KafkaAvro_Provider.h>
#include <memory>
#include <chrono>
#include <unistd.h>
#include <ctime>
#include <iomanip>
#include <Specific.hh>
#include <Encoder.hh>

#include <avroPackets/c11.h>
#include <avroPackets/c14.h>
#include <avroPackets/hk101.h>
#include <avroPackets/var102.h>
#include <avroPackets/var103.h>
#include <avroPackets/s22.h>
#include <avroPackets/cmd151.h>

using namespace inaf::oasbo::Providers;

class KafkaAvroProvider::KafkaConnectionCallback: public RdKafka::EventCb {
private:
	KafkaAvroProvider *prov;
public:
	KafkaConnectionCallback(KafkaAvroProvider *prov) {
		this->prov = prov;
	}
	void event_cb(RdKafka::Event &event) {
		if (event.err() == RdKafka::ERR__ALL_BROKERS_DOWN) {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[KafkaAvro Provider]\t"
					<< "All brokers are down, closing..." << std::endl;
			this->prov->_close();
		}
	}
};

class KafkaAvroProvider::KafkaDeliveryReportCb: public RdKafka::DeliveryReportCb {
public:
	void dr_cb(RdKafka::Message &message) {
		/* If message.err() is non-zero the message delivery failed permanently
		 * for the message. */
		if (message.err()) {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[KafkaAvro Provider]\t Message delivery failed: "
					<< message.errstr() << std::endl;
		}
	}
};

KafkaAvroProvider::KafkaAvroProvider() :
		KafkaAvroProvider("127.0.0.1", 9092, "Astri_ADAS_topic") {
}

KafkaAvroProvider::KafkaAvroProvider(std::string ip, int port,
		std::string topic_name) :
		pollThreadFlag(false), brokerIp(ip), brokerPort(port) {
	setDest(topic_name);
}

KafkaAvroProvider::~KafkaAvroProvider() {
	_close();
}

int KafkaAvroProvider::open() {
	if (isOpen())
		return 1;
	// Kafka
	RdKafka::Conf *conf = RdKafka::Conf::create(RdKafka::Conf::CONF_GLOBAL);
	std::string errstr;
	if (conf->set("bootstrap.servers",
			std::string(this->brokerIp).append(":").append(
					std::to_string(this->brokerPort)), errstr)
			!= RdKafka::Conf::CONF_OK) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[KafkaAvro Provider]\t" << errstr << std::endl;
		delete conf;
		return -1;
	}
	this->dr_cb = new KafkaDeliveryReportCb();
	if (conf->set("dr_cb", dr_cb, errstr) != RdKafka::Conf::CONF_OK) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[KafkaAvro Provider]\t" << errstr << std::endl;
		delete conf;
		return -1;
	}
	this->connection_cb = new KafkaConnectionCallback(this);
	if (conf->set("event_cb", connection_cb, errstr)
			!= RdKafka::Conf::CONF_OK) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[KafkaAvro Provider]\t" << errstr << std::endl;
		delete conf;
		return -1;
	}
	this->producer = RdKafka::Producer::create(conf, errstr);
	if (!this->producer) {
		this->producer = nullptr;
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[KafkaAvro Provider]\t" << " Failed to create producer: "
				<< errstr << std::endl;
		delete conf;
		return -1;
	}

	delete conf;
	// Check if at least one broker is available
	RdKafka::Metadata *metadata;
	RdKafka::ErrorCode err = this->producer->metadata(true, NULL, &metadata,
			5000);
	if (err != RdKafka::ERR_NO_ERROR || metadata->brokers()->size() == 0) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[KafkaAvro Provider]\t" << "No brokers available"
				<< std::endl;
		delete metadata;
		this->producer = nullptr;
		return -1;
	}

	delete metadata;
	if (!pollThreadFlag) {
		this->pollThread = std::thread(
				&KafkaAvroProvider::pollingThreadFunction, this); // start polling
		this->pollThread.detach();
		pollThreadFlag = true;
	}
	this->_open = true;
	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[KafkaAvro Provider]\t" << "Connected" << std::endl;

	return 1;
}

int KafkaAvroProvider::close() {
	if (!isOpen()) {
		return 1;
	}
	this->flush();
	this->_close();
	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[KafkaAvro Provider]\t" << "Closed" << std::endl;
	return 1;

}

void KafkaAvroProvider::flush(){
	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[KafkaAvro Provider]\t"
			<< "Flushing kafka queue in 10 seconds.." << std::endl;
	producer->flush(10 * 1000);

	if (producer->outq_len() > 0) {
		now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[KafkaAvro Provider]\t" << "" << producer->outq_len()
				<< " message(s) were not delivered" << std::endl;
	}
}

void KafkaAvroProvider::_close() {
	this->pollThreadFlag = false;
	this->_open = false;
	delete producer;
	producer = nullptr;
}


bool KafkaAvroProvider::isOpen() {
	return this->_open;
}

void KafkaAvroProvider::pollingThreadFunction() {
	producer->poll(1); // pool previous events
	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[KafkaAvro Provider]\t" << "polling thread start..."
			<< std::endl;
	while (this->pollThreadFlag) {
		if (isOpen()) {
			this->producer->poll(0);
		}
		std::this_thread::sleep_for(std::chrono::milliseconds(100)); // pool ten times at second
	}
	now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[KafkaAvro Provider]\t" << "polling thread stop..."
			<< std::endl;
}

int KafkaAvroProvider::write(Packets::BasePacket &packet) {
	return KafkaAvroProvider::write(packet, this->dest);
}

int KafkaAvroProvider::write(Packets::BasePacket &packet, std::string dest) {
	if (!isOpen()) {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[KafkaAvro Provider]\t" << "provider not open"
				<< std::endl;
		return -1;
	}
	// Reset Avro encoder state
	this->avroencoder = avro::binaryEncoder();
	auto bin_os = avro::memoryOutputStream();
	this->avroencoder->init(*bin_os);

	// Avro encoding
	this->encodeToAvro(packet);

	// FROM AVRO TO KAFKA PAYLOAD
	std::vector<uint8_t> payload;
	auto v = avro::snapshot(*bin_os);
	payload = std::move(*v);
	std::string type_subtype = std::to_string(packet[1].value())
			+ std::to_string(packet[2].value());
	int attempts = 0;
	RdKafka::ErrorCode err;
	while (attempts < 5) { //try to deliver message 5 times, then report error.
		err = this->producer->produce(
		/* Topic name */
		dest,
		/* Partition:*/
		0,
		/* Make a copy of the value */
		RdKafka::Producer::RK_MSG_COPY /* Copy payload */,
		/* Value */
		(char*) payload.data(), payload.size(),
		/* key */
		type_subtype.c_str(), type_subtype.length(),
		/* timestamp 0 = now */
		0,
		NULL);
		if (err == RdKafka::ERR_NO_ERROR)
			return 1;

		attempts += 1;

	}
	time_t now = time(nullptr);
	std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[KafkaAvro Provider]\t" << "Failed to produce to topic "
			<< dest << ": " << err << std::endl;
	return -1;
}

void KafkaAvroProvider::encodeToAvro(Packets::BasePacket &packet) {
	auto type = packet[1].value();
	auto subtype = packet[2].value();
	if (type == 2 && subtype == 2) {
		avro::encodeS22(*avroencoder, packet);
		return;
	}
	if (type == 10 && subtype == 1) {
		avro::encodeHK101(*avroencoder, packet);
		return;
	}
	if (type == 10 && subtype == 2) {
		avro::encodeVAR102(*avroencoder, packet);
		return;
	}
	if (type == 10 && subtype == 3) {
		avro::encodeVAR103(*avroencoder, packet);
		return;
	}
	if (type == 1 && subtype == 1) {
		avro::encodeC11(*avroencoder, packet);
		return;
	}
	if (type == 1 && subtype == 4) {
		avro::encodeC14(*avroencoder, packet);
		return;
	}
	if (type == 15 && subtype == 1) {
		avro::encodeCMD151(*avroencoder, packet);
		return;
	} else {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[KafkaAvro Provider]\t"
				<< " to avro error -> no known packet: " << type << "_"
				<< subtype << std::endl;
		return;
	}
}


void KafkaAvroProvider::setDest(std::string dest){
	this->dest = dest;
}

std::string KafkaAvroProvider::getDest(){
	return this->dest;
}

